import time
import math
from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.firefox.service import Service as FirefoxService
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
import pyscreenshot
import os
import shutil
import cv2 as cv
import numpy as np

## CONFIGURATION
SKIP_SNIPS = False # For debugging purposes
ZOOM_LEVEL = 18 # Zoom level of the map
TOP_LEFT_COORDS_LL = (14.8673069, 49.9896667) # Coordinates of the top-left corner of the map (longitude, latitude)
MAP_AREA_M = (1500, 1000) # Area of the map to be captured in meters (horizontal, vertical)
SNIP_REGION = (0, 60, 1880, 1020) # Defined by the top-left and bottom-right corner (x1, y1, x2, y2)
ZOOM_10_HOR_M = 140000 # Width of the map at zoom level 10 in meters (measure using the ruler in mapy.cz)
ZOOM_10_VERT_M = 65000 # Height of the map at zoom level 10 in meters (measure using the ruler in mapy.cz)
MIN_HOR_SNIP_OVERLAP = 40 # Minimum horizontal overlap between snippets in pixels
MAX_HOR_SNIP_OVERLAP = 200 # Maximum horizontal overlap between snippets in pixels
MIN_VER_SNIP_OVERLAP = 50 # Minimum vertical overlap between snippets in pixels
MAX_VER_SNIP_OVERLAP = 250 # Maximum vertical overlap between snippets in pixels
SETUP_URL = "https://mapy.cz" # URL on which the user perform the setup
SETUP_WAIT = 10 # Time in seconds to wait for the user to set up the map
MAP_SNIPS_DIR = "map_snips/" # Path to save the map snippets
FINAL_MAP_PATH = 'map.png'
MAP_LOAD_WAIT = 5 # Time in seconds to wait for the map to load before taking a screenshot
WEB_BROWSER = "firefox" # Web browser you use, either "firefox" or "chrome"
WEB_BROWSER_PATH = r"C:\Program Files\Mozilla Firefox\firefox.exe" # Path to the web browser binary (.exe)

## CONSTANTS
r_earth = 6378000 # Radius of the Earth in meters
hor_offset = ZOOM_10_HOR_M/2**(ZOOM_LEVEL-10)
vert_offset = -ZOOM_10_VERT_M/2**(ZOOM_LEVEL-10)
n_snips_cols = math.ceil(MAP_AREA_M[0]/abs(hor_offset))
n_snips_rows = math.ceil(MAP_AREA_M[1]/abs(vert_offset))

## MAP SNIPPET EXTRACTION
def extract_map_snippets():
    print(f'Collecting grid of {n_snips_rows} rows x {n_snips_cols} columns of images.')
    # Create the directory for the map snippets
    if os.path.exists(MAP_SNIPS_DIR):
        shutil.rmtree(MAP_SNIPS_DIR)
    os.makedirs(MAP_SNIPS_DIR)

    # Initialize the WebDriver based on the browser type
    if WEB_BROWSER == 'firefox':
        options = FirefoxOptions()
        if WEB_BROWSER_PATH is not None:
            options.binary_location = WEB_BROWSER_PATH
        driver = webdriver.Firefox(service=FirefoxService(), options=options)
    elif WEB_BROWSER == 'chrome':
        options = ChromeOptions()
        if WEB_BROWSER_PATH is not None:
            options.binary_location = WEB_BROWSER_PATH
        driver = webdriver.Chrome(service=ChromeService(), options=options)
    else:
        raise ValueError(f"Invalid WEB_BROWSER value: {WEB_BROWSER}")

    driver.get(SETUP_URL)
    # Wait for the page to load, let the user hide obstructing elements and change to full-screen
    time.sleep(SETUP_WAIT)

    for r in range(n_snips_rows):
        for c in range(n_snips_cols):
            # Calculate coordinates
            lon = TOP_LEFT_COORDS_LL[0] + (hor_offset*c / r_earth) * (180 / math.pi) / math.cos(TOP_LEFT_COORDS_LL[1] * math.pi/180)
            lat  = TOP_LEFT_COORDS_LL[1]  + (vert_offset*r / r_earth) * (180 / math.pi)
            # Change the map to view the coordinates at the given zoom level
            driver.get("https://mapy.cz/turisticka?l=0&x=%s&y=%s&z=%s" % (lon,lat,ZOOM_LEVEL))
            # Wait for the map to load
            time.sleep(MAP_LOAD_WAIT)
            # Take a screenshot of the map and save it
            pic = pyscreenshot.grab(bbox=SNIP_REGION)
            snip_path = os.path.join(MAP_SNIPS_DIR,"%s-%s.png" % (r,c))
            pic.save(snip_path)

    driver.quit()

## IMAGE STITCHING
def find_hor_overlap(img1, img2):
    """
    Find the best horizontal overlap between two images by comparing the right edge of the first image
    with different sections of the left edge of the second image.

    Args:
    - img1: First image (numpy array).
    - img2: Second image (numpy array).

    Returns:
    - best_hor_overlap: The best horizontal overlap in pixels.
    """
    best_img2_window_offset = 0  # Initialize the best offset for the second image
    min_diff = np.inf  # Initialize the minimum difference to infinity

    # Extract the right edge of the first image to compare
    img1_window = img1[:, -MIN_HOR_SNIP_OVERLAP:]

    # Iterate over possible offsets for the left edge of the second image
    for img2_window_offset in range(0, MAX_HOR_SNIP_OVERLAP - MIN_HOR_SNIP_OVERLAP):
        # Extract a window from the left edge of the second image
        img2_window = img2[:, img2_window_offset:(img2_window_offset + MIN_HOR_SNIP_OVERLAP)]

        # Calculate the difference between the windows of the two images
        diff = np.sum((img1_window - img2_window) ** 2)

        # Update the best offset if a smaller difference is found
        if diff < min_diff:
            min_diff = diff
            best_img2_window_offset = img2_window_offset

    # Calculate the best horizontal overlap
    best_hor_overlap = best_img2_window_offset + MIN_HOR_SNIP_OVERLAP
    return best_hor_overlap

def find_ver_overlap(img1, img2):
    """
    Find the best vertical overlap between two images by comparing the bottom edge of the first image
    with different sections of the top edge of the second image.

    Args:
    - img1: First image (numpy array).
    - img2: Second image (numpy array).

    Returns:
    - best_vert_overlap: The best vertical overlap in pixels.
    """
    best_img2_window_offset = 0  # Initialize the best offset for the second image
    min_diff = np.inf  # Initialize the minimum difference to infinity

    # Extract the bottom edge of the first image to compare
    img1_window = img1[-MIN_VER_SNIP_OVERLAP:, :]

    # Iterate over possible offsets for the top edge of the second image
    for img2_window_offset in range(0, MAX_VER_SNIP_OVERLAP - MIN_VER_SNIP_OVERLAP):
        # Extract a window from the top edge of the second image
        img2_window = img2[img2_window_offset:(img2_window_offset + MIN_VER_SNIP_OVERLAP), :]

        # Calculate the difference between the windows of the two images
        diff = np.sum((img1_window - img2_window) ** 2)

        # Update the best offset if a smaller difference is found
        if diff < min_diff:
            min_diff = diff
            best_img2_window_offset = img2_window_offset

    # Calculate the best vertical overlap
    best_vert_overlap = best_img2_window_offset + MIN_VER_SNIP_OVERLAP
    return best_vert_overlap

def stitch_images(image_grid):
    """
    Stitch the images together based on the grid and calculated overlaps.

    Args:
    - image_grid: 2D array of images.

    Returns:
    - stitched_image: Final stitched image.
    """
    
    # Initialize final stitched image
    final_stitched_image = None
    
    # Loop over each row in the grid and stitch them horizontally
    stitched_rows = []
    ver_overlaps = []
    
    for row_i,row in enumerate(image_grid):
        if row_i != 0:
            ver_overlap = find_ver_overlap(image_grid[row_i-1][0], row[0])
            ver_overlaps.append(ver_overlap)
            print(f'Rows {row_i-1}/{row_i} of column 0 overlap vertically by {ver_overlap} px')

        # Find the overlaps between neighboring images in the row
        hor_overlaps = []
        for col_i in range(1, len(row)):
            hor_overlap = find_hor_overlap(row[col_i-1], row[col_i])
            hor_overlaps.append(hor_overlap)
            print(f'Columns {col_i-1}/{col_i} of row {row_i} overlap horizontally by {hor_overlap} px')
        
        # Stitch the current row horizontally
        row_stitched = row[0]  # Start with the first image in the row
        for col_i in range(1, len(row)):
            hor_overlap = hor_overlaps[col_i-1]
            # Stitch the current image with the next one, adjusting for the overlap
            row_stitched = cv.hconcat([row_stitched, row[col_i][:, hor_overlap:]])
        
        # Add the stitched row to the list of stitched rows
        stitched_rows.append(row_stitched)
    
    # Stitch all rows vertically to form the final stitched image
    crop_correction_applied = False
    final_stitched_image = stitched_rows[0]
    for row_i in range(1, len(stitched_rows)):
        stitched_row = stitched_rows[row_i][ver_overlaps[row_i-1]:, :]
        if final_stitched_image.shape[1] != stitched_row.shape[1]:
            crop_correction_applied = True
            print("Warning: Stitched row ", row_i ," has different width than the rest of the stitched image.")
            if final_stitched_image.shape[1] > stitched_rows[row_i].shape[1]:
                final_stitched_image = final_stitched_image[:, :stitched_row.shape[1]]
            else:
                stitched_row = stitched_row[:, :final_stitched_image.shape[1]]

        final_stitched_image = cv.vconcat([final_stitched_image, stitched_row])

    if crop_correction_applied:
        print("Warning: Some of the rows had to cropped to fit the width. This indicates the image might not be stitched correctly.")

    return final_stitched_image

def get_image_grid():
    image_grid = []
    for r in range(n_snips_rows):
        row_images = []
        for c in range(n_snips_cols):
            snip_path = os.path.join(MAP_SNIPS_DIR, f"{r}-{c}.png")
            if os.path.exists(snip_path):
                image = cv.imread(snip_path)
                row_images.append(image)
            else:
                raise FileNotFoundError(f"Image not found: {snip_path}")
        image_grid.append(row_images)
    return image_grid


if not SKIP_SNIPS:
    extract_map_snippets()

image_grid = get_image_grid()

# Stitch all the captured images
stitched_map = stitch_images(image_grid)

# Save the final stitched image
cv.imwrite(FINAL_MAP_PATH, stitched_map)

print(f"Final stitched map saved at: {FINAL_MAP_PATH}")
